@@ -74,6 +74,8 @@ class Client(object):
         self.ws = None
         self.read_loop_task = None
         self.write_loop_task = None
+        self.ping_loop_task = None
+        self.ping_loop_event = self._create_event()
         self.queue = None
         self.queue_empty = None
         self.state = 'disconnected'
@@ -197,8 +199,7 @@ class Client(object):
             self._send_packet(packet.Packet(packet.CLOSE))
             self.queue.put(None)
             self.state = 'disconnecting'
-            if not abort:
-                self.queue.join()
+            self._trigger_event('disconnect')
             if self.current_transport == 'websocket':
                 self.ws.close()
             if not abort:
@@ -232,10 +233,7 @@ class Client(object):
         the Python standard library. The `start()` method on this object is
         already called by this function.
         """
-        daemon = kwargs.pop('_daemon', None)
         th = threading.Thread(target=target, args=args, kwargs=kwargs)
-        if daemon:
-            th.daemon = daemon
         th.start()
         return th
 
@@ -269,14 +267,10 @@ class Client(object):
         except ValueError:
             six.raise_from(exceptions.ConnectionError(
                 'Unexpected response from server'), None)
-        open_packet = [pkt for pkt in p.packets
-                       if pkt.packet_type == packet.OPEN]
-        if open_packet == []:
+        open_packet = p.packets[0]
+        if open_packet.packet_type != packet.OPEN:
             raise exceptions.ConnectionError(
                 'OPEN packet not returned by server')
-        if len(p.packets) > 1:  # pragma: no cover
-            self.logger.info('extra packets found in server response')
-        open_packet = open_packet[0]
         self.logger.info(
             'Polling connection accepted with ' + str(open_packet.data))
         self.sid = open_packet.data['sid']
@@ -290,6 +284,9 @@ class Client(object):
         connected_clients.append(self)
         self._trigger_event('connect')
 
+        for pkt in p.packets[1:]:
+            self._receive_packet(pkt)
+
         if 'websocket' in self.upgrades and 'websocket' in self.transports:
             # attempt to upgrade to websocket
             if self._connect_websocket(url, headers, engineio_path):
@@ -297,7 +294,7 @@ class Client(object):
                 return
 
         # start background tasks associated with this client
-        self.start_background_task(self._ping_loop, _daemon=True)
+        self.ping_loop_task = self.start_background_task(self._ping_loop)
         self.write_loop_task = self.start_background_task(self._write_loop)
         self.read_loop_task = self.start_background_task(
             self._read_loop_polling)
@@ -358,7 +355,7 @@ class Client(object):
         self.ws = ws
 
         # start background tasks associated with this client
-        self.start_background_task(self._ping_loop, _daemon=True)
+        self.ping_loop_task = self.start_background_task(self._ping_loop)
         self.write_loop_task = self.start_background_task(self._write_loop)
         self.read_loop_task = self.start_background_task(
             self._read_loop_websocket)
@@ -404,6 +401,10 @@ class Client(object):
         """Create the client's send queue."""
         return queue.Queue(), queue.Empty
 
+    def _create_event(self):
+        """Create an event."""
+        return threading.Event()
+
     def _trigger_event(self, event, *args, **kwargs):
         """Invoke an event handler."""
         run_async = kwargs.pop('run_async', False)
@@ -446,6 +447,7 @@ class Client(object):
         interval.
         """
         self.pong_received = True
+        self.ping_loop_event.clear()
         while self.state == 'connected':
             if not self.pong_received:
                 self.logger.warning(
@@ -457,7 +459,7 @@ class Client(object):
                 break
             self.pong_received = False
             self._send_packet(packet.Packet(packet.PING))
-            self.sleep(self.ping_interval)
+            self.ping_loop_event.wait(timeout=self.ping_interval)
         self.logger.info('Exiting ping task')
 
     def _read_loop_polling(self):
@@ -471,13 +473,11 @@ class Client(object):
                 self.logger.warning(
                     'Connection refused by the server, aborting')
                 self.queue.put(None)
-                self._reset()
                 break
             if r.status != 200:
                 self.logger.warning('Unexpected status code %s in server '
                                     'response, aborting', r.status)
                 self.queue.put(None)
-                self._reset()
                 break
             try:
                 p = payload.Payload(encoded_payload=r.data)
@@ -485,13 +485,22 @@ class Client(object):
                 self.logger.warning(
                     'Unexpected packet from server, aborting')
                 self.queue.put(None)
-                self._reset()
                 break
             for pkt in p.packets:
                 self._receive_packet(pkt)
 
         self.logger.info('Waiting for write loop task to end')
         self.write_loop_task.join()
+        self.logger.info('Waiting for ping loop task to end')
+        self.ping_loop_event.set()
+        self.ping_loop_task.join()
+        if self.state == 'connected':
+            self._trigger_event('disconnect')
+            try:
+                connected_clients.remove(self)
+            except ValueError:  # pragma: no cover
+                pass
+            self._reset()
         self.logger.info('Exiting read loop task')
 
     def _read_loop_websocket(self):
@@ -504,13 +513,11 @@ class Client(object):
                 self.logger.warning(
                     'WebSocket connection was closed, aborting')
                 self.queue.put(None)
-                self._reset()
                 break
             except Exception as e:
                 self.logger.info(
                     'Unexpected error "%s", aborting', str(e))
                 self.queue.put(None)
-                self._reset()
                 break
             if isinstance(p, six.text_type):  # pragma: no cover
                 p = p.encode('utf-8')
@@ -519,6 +526,16 @@ class Client(object):
 
         self.logger.info('Waiting for write loop task to end')
         self.write_loop_task.join()
+        self.logger.info('Waiting for ping loop task to end')
+        self.ping_loop_event.set()
+        self.ping_loop_task.join()
+        if self.state == 'connected':
+            self._trigger_event('disconnect')
+            try:
+                connected_clients.remove(self)
+            except ValueError:  # pragma: no cover
+                pass
+            self._reset()
         self.logger.info('Exiting read loop task')
 
     def _write_loop(self):
@@ -527,8 +544,9 @@ class Client(object):
         """
         while self.state == 'connected':
             packets = None
+            timeout = max(self.ping_interval, self.ping_timeout)
             try:
-                packets = [self.queue.get(timeout=self.ping_timeout)]
+                packets = [self.queue.get(timeout=timeout)]
             except self.queue_empty:
                 self.logger.error('packet queue is empty, aborting')
                 self._reset()
